/*
 * Copyright (C) 2004-2005 Claudio "KLaN" Ciccani <klan@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#define PACCEL  generic

/* initialize chroma components */
#define LOADUV( u, v ) \
{\
     register int _u = (u);\
     register int _v = (v);\
     cr1 = proc_table.vr[_v];\
     cr2 = proc_table.vg[_v] + proc_table.ug[_u];\
     cr3 = proc_table.ub[_u];\
}

/* perform conversion */
#define YUV2RGB( y, r, g, b ) \
{\
     register int _y;\
     _y  = proc_table.lm[(y)];\
     (r) = _y + cr1;\
     (g) = _y + cr2;\
     (b) = _y + cr3;\
}

#define CLAMP8( p )   (((uint8_t *)proc_table.clamp.p)[p])
#define CLAMP16( p )  (((uint16_t*)proc_table.clamp.p)[p])
#define CLAMP32( p )  (((uint32_t*)proc_table.clamp.p)[p])

#ifdef WORDS_BIGENDIAN
/* [YUY2: CB Y0 CR Y1] [UYVY: Y0 CB Y1 CR] */
# define YUY2_Y0( p )   (((p)&0x00ff0000)>>16)
# define YUY2_Y1( p )   (((p)&0x000000ff)    )
# define YUY2_CB( p )   (((p)&0xff000000)>>24)
# define YUY2_CR( p )   (((p)&0x0000ff00)>> 8)
# define YUY2_Y0_SHIFT  16
# define YUY2_Y1_SHIFT   0
# define YUY2_CB_SHIFT  24
# define YUY2_CR_SHIFT   8
# define UYVY_Y0_SHIFT  24
# define UYVY_Y1_SHIFT   8
# define UYVY_CB_SHIFT  16
# define UYVY_CR_SHIFT   0
#else
/* [YUY2: CR Y1 CB Y0] [UYVY: Y1 CR Y0 CB] */
# define YUY2_Y0( p )   (((p)&0x000000ff)    )
# define YUY2_Y1( p )   (((p)&0x00ff0000)>>16)
# define YUY2_CB( p )   (((p)&0x0000ff00)>> 8)
# define YUY2_CR( p )   (((p)&0xff000000)>>24)
# define YUY2_Y0_SHIFT   0
# define YUY2_Y1_SHIFT  16
# define YUY2_CB_SHIFT   8
# define YUY2_CR_SHIFT  24
# define UYVY_Y0_SHIFT   8
# define UYVY_Y1_SHIFT  24
# define UYVY_CB_SHIFT   0
# define UYVY_CR_SHIFT  16
#endif

/******************************** Begin YUY2 callbacks **********************************/

static
DFB_PFUNCTION( yuy2, yuy2 )
{
     uint32_t *S   = (uint32_t*) frame->vo_frame.base[0];
     uint32_t *D   = (uint32_t*) frame->out_plane[0];
     int       n   = frame->area/4;
     int       set = frame->mixer_set;

     do {
          register uint32_t Spix, Dpix;

          Spix = *S;
          
          if (set & MF_S) {
               Dpix  = proc_table.uv[(Spix&0x0000ff00)>> 8] << 8;
               Dpix |= proc_table.uv[(Spix&0xff000000)>>24] << 24;
          } else
               Dpix  = Spix & 0xff00ff00;

          if (set & (MF_B | MF_C)) {
               Dpix |= proc_table.lm[(Spix&0x000000ff)];
               Dpix |= proc_table.lm[(Spix&0x00ff0000)>>16] << 16;
          } else
               Dpix |= Spix & 0x00ff000ff;

          *D = Dpix;

          Spix = *(S+1);
          
          if (set & MF_S) {
               Dpix  = proc_table.uv[(Spix&0x0000ff00)>> 8] << 8;
               Dpix |= proc_table.uv[(Spix&0xff000000)>>24] << 24;
          } else
               Dpix  = Spix & 0xff00ff00;

          if (set & (MF_B | MF_C)) {
               Dpix |= proc_table.lm[(Spix&0x000000ff)];
               Dpix |= proc_table.lm[(Spix&0x00ff0000)>>16] << 16;
          } else
               Dpix |= Spix & 0x00ff000ff;

          *(D+1) = Dpix;

          S += 2;
          D += 2;
     } while (--n);
}

static
DFB_PFUNCTION( yuy2, uyvy )
{
     uint32_t *S   = (uint32_t*) frame->vo_frame.base[0];
     uint32_t *D   = (uint32_t*) frame->out_plane[0];
     int       n   = frame->area/4;
     int       set = frame->mixer_set;

     if (set) {
          do {
               register uint32_t Spix, Dpix;

               Spix = *S;
          
               if (set & MF_S) {
                    Dpix  = proc_table.uv[(Spix&0x0000ff00)>> 8];
                    Dpix |= proc_table.uv[(Spix&0xff000000)>>24] << 16;
               } else
                    Dpix  = (Spix & 0xff00ff00) >> 8;
          
               if (set & (MF_B | MF_C)) {
                    Dpix |= proc_table.lm[(Spix&0x000000ff)] << 8;
                    Dpix |= proc_table.lm[(Spix&0x00ff0000)>>16] << 24;
               } else
                    Dpix |= (Spix & 0x00ff00ff) << 8;

               *D = Dpix;

               Spix = *(S+1);
          
               if (set & MF_S) {
                    Dpix  = proc_table.uv[(Spix&0x0000ff00)>> 8];
                    Dpix |= proc_table.uv[(Spix&0xff000000)>>24] << 16;
               } else
                    Dpix  = (Spix & 0xff00ff00) >> 8;

               if (set & (MF_B | MF_C)) {
                    Dpix |= proc_table.lm[(Spix&0x000000ff)] << 8;
                    Dpix |= proc_table.lm[(Spix&0x00ff0000)>>16] << 24;
               } else
                    Dpix |= (Spix & 0x00ff00ff) << 8;

               *(D+1) = Dpix;
          
               S += 2;
               D += 2;
          } while (--n);
     } 
     else {
          do {
               register uint32_t Spix;

               Spix   = *(S+0);
               *(D+0) = ((Spix&0xff00ff00) >> 8) |
                        ((Spix&0x00ff00ff) << 8);
               
               Spix   = *(S+1);
               *(D+1) = ((Spix&0xff00ff00) >> 8) |
                        ((Spix&0x00ff00ff) << 8);

               S += 2;
               D += 2;
          } while (--n);
     }
}

static
DFB_PFUNCTION( yuy2, yv12 )
{
     uint16_t *S   = (uint16_t*) frame->vo_frame.base[0];
     uint8_t  *Dy  = frame->out_plane[0];
     uint8_t  *Du  = frame->out_plane[1];
     uint8_t  *Dv  = frame->out_plane[2];
     int       w   = frame->width;
     int       l   = frame->width/2;
     int       n   = frame->height/2;
     int       set = frame->mixer_set;

     if (set) {
          do {
               do {
                    register uint32_t Spix0, Spix1;
                    register uint32_t pu, pv;

                    Spix0 = *((uint32_t*)(S+0));
                    Spix1 = *((uint32_t*)(S+w));
                         
                    if (set & (MF_B | MF_C)) {
                         *(Dy+0)   = proc_table.lm[YUY2_Y0(Spix0)];
                         *(Dy+1)   = proc_table.lm[YUY2_Y1(Spix0)];
                         *(Dy+w+0) = proc_table.lm[YUY2_Y0(Spix1)];
                         *(Dy+w+1) = proc_table.lm[YUY2_Y1(Spix1)];
                    } else {
                         *(Dy+0)   = YUY2_Y0(Spix0);
                         *(Dy+1)   = YUY2_Y1(Spix0);
                         *(Dy+w+0) = YUY2_Y0(Spix1);
                         *(Dy+w+1) = YUY2_Y1(Spix1);
                    }

                    pu = (YUY2_CB(Spix0) + YUY2_CB(Spix1)) >> 1;
                    pv = (YUY2_CR(Spix0) + YUY2_CR(Spix1)) >> 1;
                    
                    if (set & MF_S) {
                         *Du++ = proc_table.uv[pu];
                         *Dv++ = proc_table.uv[pv];
                    } else {
                         *Du++ = pu;
                         *Dv++ = pv;
                    }

                    S  += 2;
                    Dy += 2;
               } while (--l);

               l   = w/2;
               S  += w;
               Dy += w;
          } while (--n);
     } 
     else {
          do {
               do {
                    register uint32_t Spix0, Spix1;

                    Spix0 = *((uint32_t*)(S+0));
                    Spix1 = *((uint32_t*)(S+w));
               
                    *(Dy+0)   = YUY2_Y0(Spix0);
                    *(Dy+1)   = YUY2_Y1(Spix0);
                    *(Dy+w+0) = YUY2_Y0(Spix1);
                    *(Dy+w+1) = YUY2_Y1(Spix1);
                    *Du++     = (YUY2_CB(Spix0) + YUY2_CB(Spix1)) >> 1;
                    *Dv++     = (YUY2_CR(Spix0) + YUY2_CR(Spix1)) >> 1;

                    S  += 2;
                    Dy += 2;
               } while (--l);
               
               l   = w/2;
               S  += w;
               Dy += w;
          } while (--n);
     }         
}

static
DFB_PFUNCTION( yuy2, nv12 )
{
     uint16_t *S   = (uint16_t*) frame->vo_frame.base[0];
     uint8_t  *Dy  = frame->out_plane[0];
     uint16_t *Duv = (uint16_t*) frame->out_plane[1];
     int       w   = frame->width;
     int       l   = frame->width/2;
     int       n   = frame->height/2;
     int       set = frame->mixer_set;

     if (set) {
          do {
               do {
                    register uint32_t Spix0, Spix1;
                    register uint32_t pu, pv;

                    Spix0 = *((uint32_t*)(S+0));
                    Spix1 = *((uint32_t*)(S+w));
               
                    if (set & (MF_B | MF_C)) {
                         *(Dy+0)   = proc_table.lm[YUY2_Y0(Spix0)];
                         *(Dy+1)   = proc_table.lm[YUY2_Y1(Spix0)];
                         *(Dy+w+0) = proc_table.lm[YUY2_Y0(Spix1)];
                         *(Dy+w+1) = proc_table.lm[YUY2_Y1(Spix1)];
                    } else {
                         *(Dy+0)   = YUY2_Y0(Spix0);
                         *(Dy+1)   = YUY2_Y1(Spix0);
                         *(Dy+w+0) = YUY2_Y0(Spix1);
                         *(Dy+w+1) = YUY2_Y1(Spix1);
                    }

                    pu = (YUY2_CB(Spix0) + YUY2_CB(Spix1)) >> 1;
                    pv = (YUY2_CR(Spix0) + YUY2_CR(Spix1)) >> 1;

                    if (set & MF_S)
                         *Duv++ =  proc_table.uv[pu] |
                                  (proc_table.uv[pv] << 8);
                    else
                         *Duv++ = pu | (pv << 8);

                    S  += 2;
                    Dy += 2;
               } while (--l);

               l   = w/2;
               S  += w;
               Dy += w;
          } while (--n);
     }
     else {
          do {
               do {
                    register uint32_t Spix0, Spix1;

                    Spix0 = *((uint32_t*)(S+0));
                    Spix1 = *((uint32_t*)(S+w));
               
                    *(Dy+0)   = YUY2_Y0(Spix0);
                    *(Dy+1)   = YUY2_Y1(Spix0);
                    *(Dy+w+0) = YUY2_Y0(Spix1);
                    *(Dy+w+1) = YUY2_Y1(Spix1);

                    *Duv++ =  ((YUY2_CB(Spix0)+YUY2_CB(Spix1))>>1) |
                             (((YUY2_CR(Spix0)+YUY2_CR(Spix1))>>1) << 8);

                    S  += 2;
                    Dy += 2;
               } while (--l);

               l   = w/2;
               S  += w;
               Dy += w;
          } while (--n);
     }
}

static
DFB_PFUNCTION( yuy2, nv21 )
{
     uint16_t *S   = (uint16_t*) frame->vo_frame.base[0];
     uint8_t  *Dy  = frame->out_plane[0];
     uint16_t *Dvu = (uint16_t*) frame->out_plane[1];
     int       w   = frame->width;
     int       l   = frame->width/2;
     int       n   = frame->height/2;
     int       set = frame->mixer_set;

     if (set) {
          do {
               do {
                    register uint32_t Spix0, Spix1;
                    register uint32_t pu, pv;

                    Spix0 = *((uint32_t*)(S+0));
                    Spix1 = *((uint32_t*)(S+w));
               
                    if (set & (MF_B | MF_C)) {
                         *(Dy+0)   = proc_table.lm[YUY2_Y0(Spix0)];
                         *(Dy+1)   = proc_table.lm[YUY2_Y1(Spix0)];
                         *(Dy+w+0) = proc_table.lm[YUY2_Y0(Spix1)];
                         *(Dy+w+1) = proc_table.lm[YUY2_Y1(Spix1)];
                    } else {
                         *(Dy+0)   = YUY2_Y0(Spix0);
                         *(Dy+1)   = YUY2_Y1(Spix0);
                         *(Dy+w+0) = YUY2_Y0(Spix1);
                         *(Dy+w+1) = YUY2_Y1(Spix1);
                    }

                    pu = (YUY2_CB(Spix0) + YUY2_CB(Spix1)) >> 1;
                    pv = (YUY2_CR(Spix0) + YUY2_CR(Spix1)) >> 1;

                    if (set & MF_S)
                         *Dvu++ =  proc_table.uv[pv] |
                                  (proc_table.uv[pu] << 8);
                    else
                         *Dvu++ = pv | (pu << 8);

                    S  += 2;
                    Dy += 2;
               } while (--l);

               l   = w/2;
               S  += w;
               Dy += w;
          } while (--n);
     }
     else {
          do {
               do {
                    register uint32_t Spix0, Spix1;

                    Spix0 = *((uint32_t*)(S+0));
                    Spix1 = *((uint32_t*)(S+w));
               
                    *(Dy+0)   = YUY2_Y0(Spix0);
                    *(Dy+1)   = YUY2_Y1(Spix0);
                    *(Dy+w+0) = YUY2_Y0(Spix1);
                    *(Dy+w+1) = YUY2_Y1(Spix1);

                    *Dvu++ =  ((YUY2_CR(Spix0)+YUY2_CR(Spix1))>>1) |
                             (((YUY2_CB(Spix0)+YUY2_CB(Spix1))>>1) << 8);

                    S  += 2;
                    Dy += 2;
               } while (--l);

               l   = w/2;
               S  += w;
               Dy += w;
          } while (--n);
     }
}

static
DFB_PFUNCTION( yuy2, nv16 )
{
     uint32_t *S   = (uint32_t*) frame->vo_frame.base[0];
     uint8_t  *Dy  = frame->out_plane[0];
     uint32_t *Duv = (uint32_t*) frame->out_plane[1];
     int       n   = frame->area/4;
     int       set = frame->mixer_set;

     if (set) {
          do {
               register uint32_t Spix0, Spix1;

               Spix0 = *(S+0);
               Spix1 = *(S+1);

               if (set & (MF_B | MF_C)) {
                    *(Dy+0) = proc_table.lm[YUY2_Y0(Spix0)];
                    *(Dy+1) = proc_table.lm[YUY2_Y1(Spix0)];
                    *(Dy+2) = proc_table.lm[YUY2_Y0(Spix1)];
                    *(Dy+3) = proc_table.lm[YUY2_Y1(Spix1)];
               } else {
                    *(Dy+0) = YUY2_Y0(Spix0);
                    *(Dy+1) = YUY2_Y1(Spix0);
                    *(Dy+2) = YUY2_Y0(Spix1);
                    *(Dy+3) = YUY2_Y1(Spix1);
               }

#ifdef WORDS_BIGENDIAN
               if (set & MF_S) {
                    *Duv++ = (proc_table.uv[YUY2_CB(Spix1)]      ) |
                             (proc_table.uv[YUY2_CR(Spix1)] <<  8) |
                             (proc_table.uv[YUY2_CB(Spix0)] << 16) |
                             (proc_table.uv[YUY2_CR(Spix0)] << 24);
               } else {
                    *Duv++ = (YUY2_CB(Spix1)      ) | (YUY2_CR(Spix1) <<  8) |
                             (YUY2_CB(Spix0) << 16) | (YUY2_CR(Spix0) << 24);
               }
#else
               if (set & MF_S) {
                    *Duv++ = (proc_table.uv[YUY2_CB(Spix0)]      ) |
                             (proc_table.uv[YUY2_CR(Spix0)] <<  8) |
                             (proc_table.uv[YUY2_CB(Spix1)] << 16) |
                             (proc_table.uv[YUY2_CR(Spix1)] << 24);
               } else {
                    *Duv++ = (YUY2_CB(Spix0)      ) | (YUY2_CR(Spix0) <<  8) |
                             (YUY2_CB(Spix1) << 16) | (YUY2_CR(Spix1) << 24);
               }
#endif

               S   += 2;
               Dy  += 4;
          } while (--n);
     }
     else { 
          do {
               register uint32_t Spix0, Spix1;

               Spix0 = *(S+0);
               Spix1 = *(S+1);

               *(Dy+0) = YUY2_Y0(Spix0);
               *(Dy+1) = YUY2_Y1(Spix0);
               *(Dy+2) = YUY2_Y0(Spix1);
               *(Dy+3) = YUY2_Y1(Spix1);

#ifdef WORDS_BIGENDIAN
               *Duv++  = (YUY2_CB(Spix1)      ) | (YUY2_CR(Spix1) <<  8) |
                         (YUY2_CB(Spix0) << 16) | (YUY2_CR(Spix0) << 24);
#else
               *Duv++  = (YUY2_CB(Spix0)      ) | (YUY2_CR(Spix0) <<  8) |
                         (YUY2_CB(Spix1) << 16) | (YUY2_CR(Spix1) << 24);
#endif

               S  += 2;
               Dy += 4;
          } while (--n);
     }
}

static
DFB_PFUNCTION( yuy2, rgb8 )
{
     uint32_t *S = (uint32_t*) frame->vo_frame.base[0];
     uint8_t  *D = frame->out_plane[0];
     int       n = frame->area/2;

     do {
          register uint32_t Spix;
          register int      cr1, cr2, cr3;
          register int      r, g, b;
          
          Spix = *S++;
          LOADUV( YUY2_CB(Spix), YUY2_CR(Spix) );
          
          YUV2RGB( YUY2_Y0(Spix), r, g, b );
          *(D+0) = CLAMP8(r) | CLAMP8(g) | CLAMP8(b);

          YUV2RGB( YUY2_Y1(Spix), r, g, b );
          *(D+1) = CLAMP8(r) | CLAMP8(g) | CLAMP8(b);

          D += 2;
     } while (--n);
}

static
DFB_PFUNCTION( yuy2, rgb16 )
{
     uint32_t *S = (uint32_t*) frame->vo_frame.base[0];
     uint16_t *D = (uint16_t*) frame->out_plane[0];
     int       n = frame->area/2;
     
     do {
          register uint32_t Spix;
          register int      cr1, cr2, cr3;
          register int      r, g, b;
          register uint32_t p0, p1;

          Spix = *S++;
          LOADUV( YUY2_CB(Spix), YUY2_CR(Spix) );

          YUV2RGB( YUY2_Y0(Spix), r, g, b );
          p0 = CLAMP16(r) | CLAMP16(g) | CLAMP16(b);

          YUV2RGB( YUY2_Y1(Spix), r, g, b );
          p1 = CLAMP16(r) | CLAMP16(g) | CLAMP16(b);

#ifdef WORDS_BIGENDIAN
          *((uint32_t*)D) = (p0 << 16) | p1;
#else
          *((uint32_t*)D) = p0 | (p1 << 16);
#endif

          D += 2;
     } while (--n);
}

static
DFB_PFUNCTION( yuy2, rgb24 )
{
     uint32_t *S = (uint32_t*) frame->vo_frame.base[0];
     uint8_t  *D = frame->out_plane[0];
     int       n = frame->area/2;

     do {
          register uint32_t Spix;
          register int      cr1, cr2, cr3;
          register int      r, g, b;

          Spix = *S++;
          LOADUV( YUY2_CB(Spix), YUY2_CR(Spix) );

          YUV2RGB( YUY2_Y0(Spix), r, g, b );
          *(D+0) = CLAMP8(b);
          *(D+1) = CLAMP8(g);
          *(D+2) = CLAMP8(r);

          YUV2RGB( YUY2_Y1(Spix), r, g, b );
          *(D+3) = CLAMP8(b);
          *(D+4) = CLAMP8(g);
          *(D+5) = CLAMP8(r);

          D += 6;
     } while (--n);
}

static
DFB_PFUNCTION( yuy2, rgb32 )
{
     uint32_t *S = (uint32_t*) frame->vo_frame.base[0];
     uint32_t *D = (uint32_t*) frame->out_plane[0];
     int       n = frame->area/2;

     do {
          register uint32_t Spix;
          register int      cr1, cr2, cr3;
          register int      r, g, b;

          Spix = *S++;
          LOADUV( YUY2_CB(Spix), YUY2_CR(Spix) );

          YUV2RGB( YUY2_Y0(Spix), r, g, b );
          *(D+0) = CLAMP32(r) | CLAMP32(g) | CLAMP32(b);

          YUV2RGB( YUY2_Y1(Spix), r, g, b );
          *(D+1) = CLAMP32(r) | CLAMP32(g) | CLAMP32(b);

          D += 2;
     } while (--n);
}

/******************************** End of YV12 callbacks *********************************/


/******************************** Begin YV12 callbacks **********************************/

static
DFB_PFUNCTION( yv12, yuy2 )
{
     uint8_t  *Sy  = frame->vo_frame.base[0];
     uint8_t  *Su  = frame->vo_frame.base[1];
     uint8_t  *Sv  = frame->vo_frame.base[2];
     uint16_t *D   = (uint16_t*) frame->out_plane[0];
     int       w   = frame->width;
     int       l   = frame->width/2;
     int       n   = frame->height/2;
     int       set = frame->mixer_set;

     if (set) {
          do {
               do {
                    register uint32_t Dpix;
               
                    if (set & MF_S) {
                         Dpix  = proc_table.uv[*Su++] << YUY2_CB_SHIFT;
                         Dpix |= proc_table.uv[*Sv++] << YUY2_CR_SHIFT;
                    } else {
                         Dpix  = *Su++ << YUY2_CB_SHIFT;
                         Dpix |= *Sv++ << YUY2_CR_SHIFT;
                    }
               
                    if (set & (MF_B | MF_C)) {
                         Dpix |= proc_table.lm[*(Sy+0)] << YUY2_Y0_SHIFT;
                         Dpix |= proc_table.lm[*(Sy+1)] << YUY2_Y1_SHIFT;
                    } else {
                         Dpix |= *(Sy+0) << YUY2_Y0_SHIFT;
                         Dpix |= *(Sy+1) << YUY2_Y1_SHIFT;
                    }

                    *((uint32_t*)D) = Dpix;

                    Dpix &= 0xff00ff00;

                    if (set & (MF_B | MF_C)) {
                         Dpix |= proc_table.lm[*(Sy+w+0)] << YUY2_Y0_SHIFT;
                         Dpix |= proc_table.lm[*(Sy+w+1)] << YUY2_Y1_SHIFT;
                    } else {
                         Dpix |= *(Sy+w+0) << YUY2_Y0_SHIFT;
                         Dpix |= *(Sy+w+1) << YUY2_Y1_SHIFT;
                    }

                    *((uint32_t*)(D+w)) = Dpix;

                    Sy += 2;
                    D  += 2;
               } while (--l);

               l   = w/2;
               Sy += w;
               D  += w;
          } while (--n);
     } 
     else {
          do {
               do {
                    register uint32_t Dpix;
               
                    Dpix  = *Su++   << YUY2_CB_SHIFT;
                    Dpix |= *Sv++   << YUY2_CR_SHIFT;
                    Dpix |= *(Sy+0) << YUY2_Y0_SHIFT;
                    Dpix |= *(Sy+1) << YUY2_Y1_SHIFT;

                    *((uint32_t*)D) = Dpix;

                    Dpix &= 0xff00ff00;
                    Dpix |= *(Sy+w+0) << YUY2_Y0_SHIFT;
                    Dpix |= *(Sy+w+1) << YUY2_Y1_SHIFT;

                    *((uint32_t*)(D+w)) = Dpix;

                    Sy += 2;
                    D  += 2;
               } while (--l);

               l   = w/2;
               Sy += w;
               D  += w;
          } while (--n);
     }
}

static
DFB_PFUNCTION( yv12, uyvy )
{
     uint8_t  *Sy  = frame->vo_frame.base[0];
     uint8_t  *Su  = frame->vo_frame.base[1];
     uint8_t  *Sv  = frame->vo_frame.base[2];
     uint16_t *D   = (uint16_t*) frame->out_plane[0];
     int       w   = frame->width;
     int       l   = frame->width/2;
     int       n   = frame->height/2;
     int       set = frame->mixer_set;

     if (set) {
          do {
               do {
                    register uint32_t Dpix;

                    if (set & MF_S) {
                         Dpix  = proc_table.uv[*Su++] << UYVY_CB_SHIFT;
                         Dpix |= proc_table.uv[*Sv++] << UYVY_CR_SHIFT;
                    } else {
                         Dpix  = *Su++ << UYVY_CB_SHIFT;
                         Dpix |= *Sv++ << UYVY_CR_SHIFT;
                    }

                    if (set & (MF_B | MF_C)) {
                         Dpix |= proc_table.lm[*(Sy+0)] << UYVY_Y0_SHIFT;
                         Dpix |= proc_table.lm[*(Sy+1)] << UYVY_Y1_SHIFT;
                    } else {
                         Dpix |= *(Sy+0) << UYVY_Y0_SHIFT;
                         Dpix |= *(Sy+1) << UYVY_Y1_SHIFT;
                    }

                    *((uint32_t*)D) = Dpix;

                    Dpix &= 0x00ff00ff;
               
                    if (set & (MF_B | MF_C)) {
                         Dpix |= proc_table.lm[*(Sy+w+0)] << UYVY_Y0_SHIFT;
                         Dpix |= proc_table.lm[*(Sy+w+1)] << UYVY_Y1_SHIFT;
                    } else {
                         Dpix |= *(Sy+w+0) << UYVY_Y0_SHIFT;
                         Dpix |= *(Sy+w+1) << UYVY_Y1_SHIFT;
                    }
               
                    *((uint32_t*)(D+w)) = Dpix;

                    Sy += 2;
                    D  += 2;
               } while (--l);

               l   = w/2;
               Sy += w;
               D  += w;
          } while (--n);
     }
     else {
          do {
               do {
                    register uint32_t Dpix;

                    Dpix  = *Su++   << UYVY_CB_SHIFT;
                    Dpix |= *Sv++   << UYVY_CR_SHIFT;
                    Dpix |= *(Sy+0) << UYVY_Y0_SHIFT;
                    Dpix |= *(Sy+1) << UYVY_Y1_SHIFT;

                    *((uint32_t*)D) = Dpix;

                    Dpix &= 0x00ff00ff;
                    Dpix |= *(Sy+w+0) << UYVY_Y0_SHIFT;
                    Dpix |= *(Sy+w+1) << UYVY_Y1_SHIFT;
               
                    *((uint32_t*)(D+w)) = Dpix;

                    Sy += 2;
                    D  += 2;
               } while (--l);

               l   = w/2;
               Sy += w;
               D  += w;
          } while (--n);
     }
}

static
DFB_PFUNCTION( yv12, yv12 )
{
     uint8_t *Sy  = frame->vo_frame.base[0];
     uint8_t *Su  = frame->vo_frame.base[1];
     uint8_t *Sv  = frame->vo_frame.base[2];
     uint8_t *Dy  = frame->out_plane[0];
     uint8_t *Du  = frame->out_plane[1];
     uint8_t *Dv  = frame->out_plane[2];
     int      set = frame->mixer_set;
     int      n;

     if (set & (MF_B | MF_C)) {
          n = frame->area;
          
          do {
               *Dy++ = proc_table.lm[*Sy++];
          } while (--n);
     }
     
     if (set & MF_S) {
          n = frame->area/4;
          
          do {
               *Du++ = proc_table.uv[*Su++];
               *Dv++ = proc_table.uv[*Sv++];
          } while (--n);
     }
}

static 
DFB_PFUNCTION( yv12, nv12 )
{
     uint8_t  *Sy  = frame->vo_frame.base[0];
     uint8_t  *Su  = frame->vo_frame.base[1];
     uint8_t  *Sv  = frame->vo_frame.base[2];
     uint8_t  *Dy  = frame->out_plane[0];
     uint16_t *Duv = (uint16_t*) frame->out_plane[1];
     int       set = frame->mixer_set;
     int       n;

     if (set & (MF_B | MF_C)) {
          n = frame->area;
          
          do {
               *Dy++ = proc_table.lm[*Sy++];
          } while (--n);
     }
     /* no need to copy y plane */

     n = frame->area/8;
     
     if (set & MF_S) {
          do {
               register uint32_t Dpix;
               
#ifdef WORDS_BIGENDIAN 
               Dpix  = (proc_table.uv[*(Su+1)]      ) |
                       (proc_table.uv[*(Sv+1)] <<  8) |
                       (proc_table.uv[*(Su+0)] << 16) |
                       (proc_table.uv[*(Sv+0)] << 24);
#else
               Dpix  = (proc_table.uv[*(Su+0)]      ) |
                       (proc_table.uv[*(Sv+0)] <<  8) |
                       (proc_table.uv[*(Su+1)] << 16) |
                       (proc_table.uv[*(Sv+1)] << 24);
#endif
                    
               *((uint32_t*)Duv) = Dpix;
                    
               Su  += 2;
               Sv  += 2;
               Duv += 2;
          } while (--n);
     }
     else {
          do {
               register uint32_t Dpix;

#ifdef WORDS_BIGENDIAN
               Dpix = (*(Su+1)      ) | (*(Sv+1) <<  8) |
                      (*(Su+0) << 16) | (*(Sv+0) << 24);
#else
               Dpix = (*(Su+0)      ) | (*(Sv+0) <<  8) |
                      (*(Su+1) << 16) | (*(Sv+1) << 24);
#endif
               *((uint32_t*)Duv) = Dpix;

               Su  += 2;
               Sv  += 2;
               Duv += 2;
          } while (--n);
     }
}

static
DFB_PFUNCTION( yv12, nv21 )
{
     uint8_t  *Sy  = frame->vo_frame.base[0];
     uint8_t  *Su  = frame->vo_frame.base[1];
     uint8_t  *Sv  = frame->vo_frame.base[2];
     uint8_t  *Dy  = frame->out_plane[0];
     uint16_t *Dvu = (uint16_t*) frame->out_plane[1];
     int       set = frame->mixer_set;
     int       n;

     if (set & (MF_B | MF_C)) {
          n = frame->area;
          
          do {
               *Dy++ = proc_table.lm[*Sy++];
          } while (--n);
     }
     /* no need to copy y plane */

     n = frame->area/8;
     
     if (set & MF_S) {
          do {
               register uint32_t Dpix;
               
#ifdef WORDS_BIGENDIAN 
               Dpix  = (proc_table.uv[*(Sv+1)]      ) |
                       (proc_table.uv[*(Su+1)] <<  8) |
                       (proc_table.uv[*(Sv+0)] << 16) |
                       (proc_table.uv[*(Su+0)] << 24);
#else
               Dpix  = (proc_table.uv[*(Sv+0)]      ) |
                       (proc_table.uv[*(Su+0)] <<  8) |
                       (proc_table.uv[*(Sv+1)] << 16) |
                       (proc_table.uv[*(Su+1)] << 24);
#endif
                    
               *((uint32_t*)Dvu) = Dpix;
                    
               Su  += 2;
               Sv  += 2;
               Dvu += 2;
          } while (--n);
     }
     else {
          do {
               register uint32_t Dpix;

#ifdef WORDS_BIGENDIAN
               Dpix = (*(Sv+1)      ) | (*(Su+1) <<  8) |
                      (*(Sv+0) << 16) | (*(Su+0) << 24);
#else
               Dpix = (*(Sv+0)      ) | (*(Su+0) <<  8) |
                      (*(Sv+1) << 16) | (*(Su+1) << 24);
#endif
               *((uint32_t*)Dvu) = Dpix;

               Su  += 2;
               Sv  += 2;
               Dvu += 2;
          } while (--n);
     }
}

static
DFB_PFUNCTION( yv12, nv16 )
{ 
     uint8_t  *Sy  = frame->vo_frame.base[0];
     uint8_t  *Su  = frame->vo_frame.base[1];
     uint8_t  *Sv  = frame->vo_frame.base[2];
     uint8_t  *Dy  = frame->out_plane[0];
     uint16_t *Duv = (uint16_t*) frame->out_plane[1];
     int       w   = frame->width/2;
     int       l   = frame->width/4;
     int       set = frame->mixer_set;
     int       n;

     if (set & (MF_B | MF_C)) {
          n = frame->area;
          
          do {
               *Dy++ = proc_table.lm[*Sy++];
          } while (--n);
     }
     /* no need to copy y plane */

     n = frame->height/2;
     
     if (set & MF_S) {
          do {
               do {
                    register uint32_t Dpix;

#ifdef WORDS_BIGENDIAN 
                    Dpix  = (proc_table.uv[*(Su+1)]      ) |
                            (proc_table.uv[*(Sv+1)] <<  8) |
                            (proc_table.uv[*(Su+0)] << 16) |
                            (proc_table.uv[*(Sv+0)] << 24);
#else
                    Dpix  = (proc_table.uv[*(Su+0)]      ) |
                            (proc_table.uv[*(Sv+0)] <<  8) |
                            (proc_table.uv[*(Su+1)] << 16) |
                            (proc_table.uv[*(Sv+1)] << 24);
#endif
                    *((uint32_t*)(Duv+0)) = Dpix;
                    *((uint32_t*)(Duv+w)) = Dpix;
                    
                    Su  += 2;
                    Sv  += 2;
                    Duv += 2;
               } while (--l);

               l    = w/2;
               Duv += w;
          } while (--n);
     }
     else {
          do {
               do {
                    register uint32_t Dpix;

#ifdef WORDS_BIGENDIAN
                    Dpix = (*(Su+1)      ) | (*(Sv+1) <<  8) |
                           (*(Su+0) << 16) | (*(Sv+0) << 24);
#else
                    Dpix = (*(Su+0)      ) | (*(Sv+0) <<  8) |
                           (*(Su+1) << 16) | (*(Sv+1) << 24);
#endif
                    *((uint32_t*)(Duv+0)) = Dpix;
                    *((uint32_t*)(Duv+w)) = Dpix;

                    Su  += 2;
                    Sv  += 2;
                    Duv += 2;
               } while (--l);
                
               l    = w/2;
               Duv += w;
          } while (--n);
     }
}

static
DFB_PFUNCTION( yv12, rgb8 )
{
     uint8_t *Sy = frame->vo_frame.base[0];
     uint8_t *Su = frame->vo_frame.base[1];
     uint8_t *Sv = frame->vo_frame.base[2];
     uint8_t *D  = frame->out_plane[0];
     int      w  = frame->width;
     int      l  = frame->width/2;
     int      n  = frame->height/2;

     do {
          do {
               register int cr1, cr2, cr3;
               register int r, g, b;

               LOADUV( *Su++, *Sv++ );

               YUV2RGB( *(Sy+0), r, g, b );
               *(D+0) = CLAMP8(r) | CLAMP8(g) | CLAMP8(b);

               YUV2RGB( *(Sy+1), r, g, b );
               *(D+1) = CLAMP8(r) | CLAMP8(g) | CLAMP8(b);

               YUV2RGB( *(Sy+w), r, g, b );
               *(D+w) = CLAMP8(r) | CLAMP8(g) | CLAMP8(b);

               YUV2RGB( *(Sy+w+1), r, g, b );
               *(D+w+1) = CLAMP8(r) | CLAMP8(g) | CLAMP8(b);

               Sy += 2;
               D  += 2;
          } while (--l);

          l   = w/2;
          Sy += w;
          D  += w;
     } while (--n);

}

static
DFB_PFUNCTION( yv12, rgb16 )
{
     uint8_t  *Sy = frame->vo_frame.base[0];
     uint8_t  *Su = frame->vo_frame.base[1];
     uint8_t  *Sv = frame->vo_frame.base[2];
     uint16_t *D  = (uint16_t*) frame->out_plane[0];
     int       w  = frame->width;
     int       l  = frame->width/2;
     int       n  = frame->height/2;
     
     do {
          do {
               register int      cr1, cr2, cr3;
               register int      r, g, b;
               register uint32_t p0, p1;

               LOADUV( *Su++, *Sv++ );

               YUV2RGB( *(Sy+0), r, g, b );
               p0 = CLAMP16(r) | CLAMP16(g) | CLAMP16(b);

               YUV2RGB( *(Sy+1), r, g, b );
               p1 = CLAMP16(r) | CLAMP16(g) | CLAMP16(b);

#ifdef WORDS_BIGENDIAN
               *((uint32_t*)D) = (p0 << 16) | p1;
#else
               *((uint32_t*)D) = p0 | (p1 << 16);
#endif

               YUV2RGB( *(Sy+w), r, g, b );
               p0 = CLAMP16(r) | CLAMP16(g) | CLAMP16(b);
          
               YUV2RGB( *(Sy+w+1), r, g, b );
               p1 = CLAMP16(r) | CLAMP16(g) | CLAMP16(b);

#ifdef WORDS_BIGENDIAN
               *((uint32_t*)(D+w)) = (p0 << 16) | p1;
#else
               *((uint32_t*)(D+w)) = p0 | (p1 << 16);
#endif

               Sy += 2;
               D  += 2;
          } while (--l);

          l   = w/2;
          Sy += w;
          D  += w;
     } while (--n);
}

static
DFB_PFUNCTION( yv12, rgb24 )
{
     uint8_t *Sy = frame->vo_frame.base[0];
     uint8_t *Su = frame->vo_frame.base[1];
     uint8_t *Sv = frame->vo_frame.base[2];
     uint8_t *D  = frame->out_plane[0];
     int      w  = frame->width;
     int      l  = frame->width/2;
     int      n  = frame->height/2;

     do {
          do {
               register int cr1, cr2, cr3;
               register int r, g, b;

               LOADUV( *Su++, *Sv++ );

               YUV2RGB( *(Sy+0), r, g, b );
               *(D+0) = CLAMP8(b);
               *(D+1) = CLAMP8(g);
               *(D+2) = CLAMP8(r);

               YUV2RGB( *(Sy+1), r, g, b );
               *(D+3) = CLAMP8(b);
               *(D+4) = CLAMP8(g);
               *(D+5) = CLAMP8(r);

               YUV2RGB( *(Sy+w), r, g, b );
               *(D+w+0) = CLAMP8(b);
               *(D+w+1) = CLAMP8(g);
               *(D+w+2) = CLAMP8(r);

               YUV2RGB( *(Sy+w+1), r, g, b );
               *(D+w+3) = CLAMP8(b);
               *(D+w+4) = CLAMP8(g);
               *(D+w+5) = CLAMP8(r);

               Sy += 2;
               D  += 6;
          } while (--l);

          l   = w/2;
          Sy += w;
          D  += w*3;
     } while (--n);
}

static
DFB_PFUNCTION( yv12, rgb32 )
{
     uint8_t  *Sy = frame->vo_frame.base[0];
     uint8_t  *Su = frame->vo_frame.base[1];
     uint8_t  *Sv = frame->vo_frame.base[2];
     uint32_t *D  = (uint32_t*) frame->out_plane[0];
     int       w  = frame->width;
     int       l  = frame->width/2;
     int       n  = frame->height/2;

     do {
          do {
               register int cr1, cr2, cr3;
               register int r, g, b;

               LOADUV( *Su++, *Sv++ );

               YUV2RGB( *(Sy+0), r, g, b );
               *(D+0) = CLAMP32(r) | CLAMP32(g) | CLAMP32(b);

               YUV2RGB( *(Sy+1), r, g, b );
               *(D+1) = CLAMP32(r) | CLAMP32(g) | CLAMP32(b);

               YUV2RGB( *(Sy+w), r, g, b );
               *(D+w) = CLAMP32(r) | CLAMP32(g) | CLAMP32(b);

               YUV2RGB( *(Sy+w+1), r, g, b );
               *(D+w+1) = CLAMP32(r) | CLAMP32(g) | CLAMP32(b);

               Sy += 2;
               D  += 2;
          } while (--l);

          l   = w/2;
          Sy += w;
          D  += w;
     } while (--n);
}

/******************************** End of YV12 callbacks *********************************/

#undef PACCEL

