/*
 * Copyright (C) 2004-2005 Claudio "KLaN" Ciccani <klan@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  written by Claudio Ciccani <klan82@cheapnet.it> with the contribute
 *  of Kristof Pelckmans <kristof.pelckmans@antwerpen.be>
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "df_xine.h"


void
dfx_playlist_append( DFXCore *this, const char *mrl )
{
     DFXMedia *media = &this->media;
     DFXMrl   *cur;

     cur = (DFXMrl*) DX_ALLOC( sizeof( DFXMrl ) );

     cur->mrl = DX_STRDUP( mrl );

     if (media->first) {
          cur->prev         = media->last;
          media->last->next = cur;
          media->last       = cur;
     } else {
          media->first = cur;
          media->last  = cur;
          media->cur   = cur;
     }
}

void
dfx_playlist_discard( DFXCore *this )
{
     DFXMedia *media = &this->media;
     DFXMrl   *cur   = media->cur;

     if (cur->prev && cur->next) {
          cur->next->prev  = cur->prev;
          cur->prev->next  = cur->next;
          media->cur       = cur->next;
     } else
     if (cur->prev) { /* end of playlist */
          cur->prev->next = NULL;
          media->cur      = media->first;
     } else
     if (cur->next) { /* start of playlist */
          cur->next->prev = NULL;
          media->cur      = cur->next;
     } else {
          media->cur   = NULL;
          media->first = NULL;
          media->last  = NULL;
     }

     free( cur->mrl );
     free( cur );
}

void
dfx_playlist_skip( DFXCore *this, int rel )
{
     DFXMedia *media = &this->media;
     DFXMrl   *cur   = media->cur;
     int       n     = abs( rel );
     
     if (!cur || !rel)
          return;

     do {
          if (rel > 0)
               cur = cur->next ? : media->first;
          else
               cur = cur->prev ? : media->last;
     } while (--n);

     media->cur = cur;
}

void
dfx_playlist_free( DFXCore *this )
{
     DFXMrl *cur  = this->media.first;
     DFXMrl *next;

     while (cur) {
          next = cur->next;
          free( cur->mrl );
          free( cur );
          cur = next;
     }
}
